#  Copyright 2021 Google LLC
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import jax
from optax._src.linear_algebra import global_norm


def default_loss_and_grads(train_rng, state, batch, loss_fn):
    def compute_loss(params):
        return loss_fn(
            train_rng=train_rng,
            state=state,
            params=params,
            batch=batch,
            is_training=True)

    grad_fn = jax.value_and_grad(compute_loss)
    return grad_fn(state.params)


def default_train_step(train_rng,
                       state,
                       batch,
                       loss_fn,
                       lr_schedule_fn,
                       under_pmap):
    loss, grads = default_loss_and_grads(
        train_rng=train_rng, state=state, batch=batch, loss_fn=loss_fn)

    if under_pmap:
        grads = jax.lax.pmean(grads, 'batch')

    new_state = state.apply_gradients(grads=grads)

    metrics = {
        'loss': loss,
        'step': state.step,
        'grad_norm': global_norm(grads)
    }
    if lr_schedule_fn is not None:
        metrics.update({'lr': lr_schedule_fn(state.step)})

    if under_pmap:
        metrics = jax.lax.pmean(metrics, axis_name='batch')

    return new_state, metrics


def default_eval_step(state, batch, loss_fn, under_pmap):
    loss = loss_fn(
        train_rng=jax.random.PRNGKey(0),
        state=state,
        params=state.params,
        batch=batch,
        is_training=False)

    metrics = {'loss': loss}
    if under_pmap:
        metrics = jax.lax.pmean(metrics, axis_name='batch')

    return metrics



#  Copyright 2021 Google LLC
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np
import jax
from jax.experimental.pjit import pjit
from jax.sharding import Mesh
from jax.sharding import PartitionSpec as P
from flax.core.frozen_dict import FrozenDict, unfreeze
from flax.traverse_util import flatten_dict
import optax

from .partition_utils import set_partitions


def get_mesh(n_model_shards):
    if n_model_shards == 1:
        return None

    assert jax.device_count() % n_model_shards == 0

    mesh_devices = np.array(jax.devices()).reshape(
        jax.device_count() // n_model_shards, n_model_shards)
    mesh = Mesh(mesh_devices, ('dp', 'mp'))

    return mesh


class ShapeDtypeStruct:
    __slots__ = ["shape", "dtype"]

    def __init__(self, shape, dtype):
        self.shape = shape
        self.dtype = dtype


def get_param_spec(params, params_sharding_rules):
    return set_partitions(unfreeze(params), params_sharding_rules)


def shard_params(params, params_spec, mesh):
    shard_fn = pjit(
        lambda x: x, in_shardings=(None,), out_shardings=params_spec)

    with mesh:
        return shard_fn(params)


def shard_params_and_opt_state(
        params, params_spec, mesh, optimizer, init_opt_state=None):
    def init_fn(params_):
        if init_opt_state is None:
            opt_state_ = optimizer.init(params_)
        else:
            opt_state_ = init_opt_state

        return opt_state_, params_

    def get_opt_spec(x):
        if isinstance(x, (dict, FrozenDict,)):
            return params_spec
        return None

    params_shapes = jax.tree_util.tree_map(
        lambda x: ShapeDtypeStruct(x.shape, x.dtype), params)
    state_shapes = jax.eval_shape(init_fn, params_shapes)

    opt_state_spec, _ = jax.tree_util.tree_map(
        get_opt_spec, state_shapes,
        is_leaf=lambda x: isinstance(x, (dict, FrozenDict, optax.EmptyState,)))

    p_get_initial_state = pjit(
        init_fn,
        in_shardings=(params_spec,),
        out_shardings=(opt_state_spec, params_spec))

    with mesh:
        opt_state, params = p_get_initial_state(params)

    return params, opt_state, opt_state_spec


def get_sharding_rules(params, mesh_model_shards, investigate_depth=2):
    sharding_rules = {
        ('(bias|scale)',): None,
        ('embedding',): P('mp', None),
    }

    last_dense_mp_dim = None
    flat_params = flatten_dict(params)
    for key in sorted(flat_params.keys(), key=lambda t: (len(t), t)):
        param = flat_params[key]

        rule_key = key[-investigate_depth:]

        if key[-1] in ['bias', 'scale']:
            assert len(param.shape) == 1

        elif key[-1] == 'embedding':
            assert len(param.shape) == 2
            if param.shape[0] % mesh_model_shards != 0:
                sharding_rules[('embedding',)] = P(None, 'mp')

        else:
            if len(param.squeeze().shape) == 1:
                sharding_rules[rule_key] = None

            elif rule_key in sharding_rules:
                for dim_size, rule_str in zip(
                        param.shape, sharding_rules[rule_key]):
                    assert rule_str != 'mp' or dim_size % mesh_model_shards == 0

            elif under_attention(key) and rule_key[0][0] == 'o':
                sharding_rules[rule_key] = P('mp', None)

            elif under_attention(key) and rule_key[0][0] in ['q', 'k', 'v']:
                sharding_rules[rule_key] = P(None, 'mp')

            elif under_attention(key) and rule_key[0][-1] == 'o':
                sharding_rules[rule_key] = P('mp', None)

            elif under_attention(key) and rule_key[0][-1] in ['q', 'k', 'v']:
                sharding_rules[rule_key] = P(None, 'mp')

            else:
                rule_tuple = [None for _ in range(len(param.shape))]
                for dim in range(-1, -len(param.shape) - 1, -1):
                    if dim != last_dense_mp_dim and \
                            param.shape[dim] % mesh_model_shards == 0:
                        last_dense_mp_dim = dim
                        rule_tuple[dim] = 'mp'
                        break
                if all([t is None for t in rule_tuple]):
                    if last_dense_mp_dim is not None and \
                            param.shape[last_dense_mp_dim] % \
                            mesh_model_shards == 0:
                        rule_tuple[last_dense_mp_dim] = 'mp'

                sharding_rules[rule_key] = P(*rule_tuple)

    return list(sharding_rules.items())


def under_attention(flat_param_key):
    for key in flat_param_key:
        if 'attention' in key.lower() or 'attn' in key.lower():
            return True
    return False



#  Copyright 2021 Google LLC
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from glob import glob
import tqdm

from .dataset import Dataset


class JsonlDataset(Dataset):
    def __init__(self, data_dir):
        self._split_filenames = {}
        for filename in glob(f'{data_dir}/*.jsonl'):
            split = filename.split('/')[-1][:-len('.jsonl')]
            self._split_filenames[split] = filename

    def __getitem__(self, split):
        examples = []
        for line in tqdm.tqdm(open(self._split_filenames[split]),
                              desc=f'loading {split} examples'):
            examples.append(json.loads(line))

        return examples



#  Copyright 2021 Google LLC
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import json
import jax


def get_logger(verbose, workdir):
    logger = logging.getLogger('redco')

    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(
        fmt="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S"))
    logger.addHandler(handler)

    if workdir is not None:
        handler = logging.FileHandler(filename=f'{workdir}/log.txt')
        handler.setFormatter(logging.Formatter(
            fmt="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
            datefmt="%m/%d/%Y %H:%M:%S"))
        logger.addHandler(handler)

    logger.propagate = False

    if verbose and jax.process_index() == 0:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.ERROR)

    for handler in logging.root.handlers:
        logging.root.removeHandler(handler)

    return logger


def log_info(info, title, logger, summary_writer, step):
    info = str(info)

    if title is not None:
        if step is not None:
            title_ = f'{title} (step {step})'
        else:
            title_ = title
            step = 0

        if summary_writer is not None:
            summary_writer.text(title, info.replace('\n', '\n\n'), step=step)

        max_len = max(max([len(t) for t in info.split('\n')]), len(title_) + 4)

        logger.info('=' * max_len)
        logger.info(f'### {title_}')
        logger.info('-' * max_len)
        for t in info.split('\n'):
            logger.info(t)
        logger.info('=' * max_len)

    else:
        logger.info(info)


def save_outputs(outputs, workdir, desc, logger, summary_writer, step):
    outputs = jax.tree_util.tree_map(str, outputs)
    json.dump(outputs, open(f'{workdir}/outputs_{desc}.json', 'w'), indent=4)
    logger.info(
        f'Outputs ({desc}) has been saved into {workdir}/outputs_{desc}.json.')

    if summary_writer is not None:
        samples_str = json.dumps(outputs[:10], indent=4).replace('\n', '\n\n')
        summary_writer.text('outputs', samples_str, step=step)

    return f'{workdir}/outputs_{desc}.json'


#  Copyright 2021 Google LLC
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  adapted from https://github.com/google-research/google-research/blob/fce923e9dad97cd67492c2a65b9ecdc4b2495204/flax_models/t5x/partitions.py
"""Utilities for constructing PyTrees of PartitionSpecs."""

import re

from flax.core.frozen_dict import freeze
from flax.traverse_util import flatten_dict, unflatten_dict


# Sentinels
_unmatched = object()


def _match(qs, ks):
    """Return True if regexes in qs match any window of strings in tuple ks."""
    # compile regexes and force complete match
    qts = tuple(map(lambda x: re.compile(x + "$"), qs))
    for i in range(len(ks) - len(qs) + 1):
        matches = [x.match(y) for x, y in zip(qts, ks[i:])]
        if matches and all(matches):
            return True
    return False


def _replacement_rules(rules):
    def replace(key, val):
        for rule, replacement in rules:
            if _match(rule, key):
                return replacement
        return val

    return replace


def set_partitions(in_dict, rules):
    replace = _replacement_rules(rules)
    initd = {k: _unmatched for k in flatten_dict(in_dict)}
    result = {k: replace(k, v) for k, v in initd.items()}
    assert _unmatched not in result.values(), "Incomplete partition spec."
    return freeze(unflatten_dict(result))



#  Copyright 2021 Google LLC
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
import jax
import jax.numpy as jnp
from jax.experimental import multihost_utils
from flax.jax_utils import replicate, unreplicate
from flax.training.common_utils import shard_prng_key
from flax.core.frozen_dict import freeze, unfreeze
from flax.serialization import (
    msgpack_serialize, msgpack_restore, from_state_dict, to_state_dict)

from .data_utils import get_host_examples, get_data_batches
from .opt_utils import get_lr_schedule_fn
from .log_utils import get_logger, log_info, save_outputs
from .model_parallel_utils.mesh_utils import (
    get_mesh,
    shard_params_and_opt_state,
    shard_params,
    get_param_spec,
    get_sharding_rules)


class Deployer:
    def __init__(self,
                 jax_seed,
                 n_model_shards=1,
                 verbose=True,
                 workdir=None,
                 n_processes=1,
                 host0_address=None,
                 host0_port=None,
                 process_id=None,
                 n_local_devices=None,
                 run_tensorboard=False,
                 run_wandb=False):
        if n_processes > 1:
            local_device_ids = None if n_local_devices is None \
                else list(range(n_local_devices))

            jax.distributed.initialize(
                coordinator_address=f'{host0_address}:{host0_port}',
                num_processes=n_processes,
                process_id=process_id,
                local_device_ids=local_device_ids)

            print(f'process_id: {jax.process_index()} / {jax.process_count()}')
            print(f'devices: {jax.local_device_count()} / {jax.device_count()}')

        if workdir is not None:
            os.makedirs(workdir, exist_ok=True)

        self._verbose = verbose
        self._workdir = workdir
        self._logger = get_logger(verbose=verbose, workdir=workdir)

        if run_wandb:
            import wandb
            self._wandb_log_fn = wandb.log
        else:
            self._wandb_log_fn = None

        if run_tensorboard and jax.process_index() == 0:
            from flax.metrics import tensorboard
            self._summary_writer = tensorboard.SummaryWriter(workdir)
        else:
            self._summary_writer = None

        self._rng = jax.random.PRNGKey(seed=jax_seed)
        self._mesh = get_mesh(n_model_shards=n_model_shards)

    def process_batch_size(self, per_device_batch_size):
        if self._mesh is None:
            batch_size = per_device_batch_size * jax.local_device_count()
            global_batch_size = batch_size * jax.process_count()
        else:
            batch_size = per_device_batch_size * self._mesh.shape['dp']
            global_batch_size = batch_size

        return batch_size, global_batch_size

    def get_model_input_batches(self,
                                examples,
                                per_device_batch_size,
                                collate_fn,
                                shuffle,
                                shuffle_rng,
                                desc,
                                is_train=False,
                                accumulate_grad_batches=None):
        batch_size, global_batch_size = self.process_batch_size(
            per_device_batch_size=per_device_batch_size)

        examples = get_host_examples(
            examples=examples,
            global_batch_size=global_batch_size,
            shuffle=shuffle,
            shuffle_rng=shuffle_rng,
            mesh=self._mesh)

        if not is_train:
            desc = f'{desc} (global_batch_size = {global_batch_size})'
        elif accumulate_grad_batches is None:
            desc = f'{desc} (global_micro_batch_size = {global_batch_size})'
        else:
            desc = (f'{desc} ('
                    f'global_micro_batch_size = {global_batch_size}, '
                    f'accumulate_grad_batches = {accumulate_grad_batches})')

        return get_data_batches(
            examples=examples,
            batch_size=batch_size,
            collate_fn=collate_fn,
            do_shard=(self.mesh is None),
            desc=desc,
            verbose=self._verbose)

    def process_batch_preds(self, batch_preds_with_idxes):
        if self._mesh is None:
            batch_preds_with_idxes = jax.tree_util.tree_map(
                lambda x: x[0], batch_preds_with_idxes)

            batch_preds = batch_preds_with_idxes['raw_preds']
            idxes = batch_preds_with_idxes['__idx__']

            preds = jax.tree_util.tree_map(
                lambda t: t.reshape((t.shape[0] * t.shape[1],) + t.shape[2:]),
                batch_preds)
            idxes = idxes.reshape(-1)
            idxes_argsort = jnp.argsort(idxes, axis=None)

            return jax.tree_util.tree_map(lambda t: t[idxes_argsort], preds)
        else:
            return batch_preds_with_idxes['raw_preds']

    def process_to_run_model(self, x, is_prng_key=False):
        if self._mesh is None:
            if is_prng_key:
                return shard_prng_key(x)
            else:
                return replicate(x)
        else:
            return x

    def process_to_deliver(self, x):
        if self._mesh is None:
            return unreplicate(x)
        else:
            return x

    def get_lr_schedule_fn(self,
                           train_size,
                           per_device_batch_size,
                           n_epochs,
                           learning_rate,
                           schedule_type='linear',
                           warmup_rate=0.,
                           warmup_steps=None,
                           init_learning_rate=0.,
                           end_learning_rate=0.):
        _, global_batch_size = self.process_batch_size(
            per_device_batch_size=per_device_batch_size)
        total_train_steps = n_epochs * (train_size // global_batch_size)

        if warmup_steps is None:
            warmup_steps = int(total_train_steps * warmup_rate)

        return get_lr_schedule_fn(
            schedule_type=schedule_type,
            total_train_steps=total_train_steps,
            warmup_steps=warmup_steps,
            init_learning_rate=init_learning_rate,
            learning_rate=learning_rate,
            end_learning_rate=end_learning_rate)

    def get_sharding_rules(self, params):
        if self._mesh is None:
            return None
        else:
            sharding_rules = get_sharding_rules(
                params=params, mesh_model_shards=self._mesh.shape['mp'])

            self.log_info(
                info='\n'.join([f'{t}' for t in sharding_rules]),
                title='Sharding rules')

            return sharding_rules

    def get_params_spec(self, params, params_sharding_rules):
        return get_param_spec(
            params=params, params_sharding_rules=params_sharding_rules)

    def shard_params(self, params, params_spec):
        return shard_params(
            params=params, params_spec=params_spec, mesh=self._mesh)

    def shard_params_and_opt_state(
            self, params, params_spec, optimizer, init_opt_state):
        return shard_params_and_opt_state(
            params=params,
            params_spec=params_spec,
            mesh=self._mesh,
            optimizer=optimizer,
            init_opt_state=init_opt_state)

    def run_model_step(self, step_fn, input_args):
        if self._mesh is None:
            return step_fn(*input_args)
        else:
            with self._mesh:
                return step_fn(*input_args)

    def gen_rng(self):
        self._rng, new_rng = jax.random.split(self._rng)
        return new_rng

    def log_info(self, info, title=None, step=None):
        if jax.process_index() == 0:
            log_info(
                info=info,
                title=title,
                logger=self._logger,
                summary_writer=self._summary_writer,
                step=step)

    def log_metrics(self, metrics, step):
        if self._summary_writer is not None:
            for metric_name, value in metrics.items():
                self._summary_writer.scalar(metric_name, value, step=step)

        if self._wandb_log_fn is not None and jax.process_index() == 0:
            self._wandb_log_fn(metrics, step)

    def save_outputs(self, outputs, desc, step):
        if self._workdir is not None and jax.process_index() == 0:
            save_outputs(
                workdir=self._workdir,
                outputs=outputs,
                desc=desc,
                step=step,
                logger=self._logger,
                summary_writer=self._summary_writer)

    def load_params(self, filepath):
        with jax.default_device(jax.devices('cpu')[0]):
            params = msgpack_restore(open(filepath, 'rb').read())
            params = jax.tree_util.tree_map(jnp.asarray, params)

        self.log_info(f'params loaded from {filepath}')
        return params

    def load_opt_state(self, ckpt_dir, desc, params, optimizer):
        filepath = f'{ckpt_dir}/opt_state_{desc}.msgpack'
        if not os.path.exists(filepath):
            self.log_info(f'Skip loading opt_state (No file {filepath})')
            return None

        with jax.default_device(jax.devices('cpu')[0]):
            opt_state = msgpack_restore(open(filepath, 'rb').read())
            opt_state = from_state_dict(
                target=optimizer.init(freeze(params)), state=opt_state)

        self.log_info(f'opt_state loaded from {filepath}.')
        return opt_state

    def save_params(self, params, ckpt_dir, desc):
        if self._mesh is None:
            params = unreplicate(params)
        else:
            with jax.default_device(jax.devices('cpu')[0]):
                params = multihost_utils.process_allgather(params)

        if jax.process_index() == 0:
            filepath = f'{ckpt_dir}/params_{desc}.msgpack'
            open(filepath, "wb").write(msgpack_serialize(unfreeze(params)))
            self.log_info(f'params saved into {filepath}')

    def save_opt_state(self, opt_state, ckpt_dir, desc):
        if self._mesh is None:
            opt_state = unreplicate(opt_state)
        else:
            with jax.default_device(jax.devices('cpu')[0]):
                opt_state = multihost_utils.process_allgather(opt_state)

        if jax.process_index() == 0:
            filepath = f'{ckpt_dir}/opt_state_{desc}.msgpack'
            opt_state = to_state_dict(opt_state)
            open(filepath, "wb").write(msgpack_serialize(unfreeze(opt_state)))
            self.log_info(f'opt_state saved into {filepath}')

    def save_rng(self, ckpt_dir, desc):
        if jax.process_index() == 0:
            jnp.save(f'{ckpt_dir}/rng_{desc}.npy', self._rng)

    @property
    def mesh(self):
        return self._mesh

    @property
    def workdir(self):
        return self._workdir



#  Copyright 2021 Google LLC
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from functools import partial
import json
import numpy as np
import jax
import jax.numpy as jnp
from jax.experimental.pjit import pjit
from jax.sharding import PartitionSpec as P
from flax.jax_utils import replicate
from flax.training import train_state
from flax.traverse_util import flatten_dict
from flax.core.frozen_dict import freeze

from .utils import default_train_step, default_eval_step
from ..predictors import Predictor


class Trainer:
    def __init__(self,
                 deployer,
                 collate_fn,
                 apply_fn,
                 loss_fn,
                 params,
                 optimizer,
                 lr_schedule_fn=None,
                 accumulate_grad_batches=None,
                 params_sharding_rules=None):
        self._deployer = deployer
        self._collate_fn = collate_fn
        self._apply_fn = apply_fn
        self._loss_fn = loss_fn
        self._optimizer = optimizer
        self._lr_schedule_fn = lr_schedule_fn
        self._accumulate_grad_batches = accumulate_grad_batches
        self._params_sharding_rules = params_sharding_rules

        self._state = None
        self._state_spec = None
        self._p_train_step = None
        self._p_eval_step = None

        n_params = sum([param.size for param in flatten_dict(params).values()])
        self._deployer.log_info(f'{n_params:,}', title='Training parameters')

        last_ckpt_info_path = \
            f'{self.workdir}/ckpts/last_ckpt_info.json'
        if os.path.exists(last_ckpt_info_path):
            self._last_ckpt_info = json.load(open(last_ckpt_info_path))
            desc = self._last_ckpt_info['last_desc']
            params = self._deployer.load_params(
                filepath=f'{self.workdir}/ckpts/params_{desc}.msgpack')
            opt_state = self._deployer.load_opt_state(
                ckpt_dir=f'{self.workdir}/ckpts',
                desc=desc,
                params=params,
                optimizer=optimizer)

            self.create_train_state(
                apply_fn=self._apply_fn,
                params=params,
                optimizer=self._optimizer,
                step=self._last_ckpt_info['last_step'],
                init_opt_state=opt_state)

            self._deployer._rng = jnp.load(
                f'{self.workdir}/ckpts/rng_{desc}.npy')

            self._deployer.log_info(
                'loaded last_ckpt \"{last_desc}\",'
                ' last_step={last_step},'
                ' last_epoch_idx={last_epoch_idx}'.format(
                    **self._last_ckpt_info))
        else:
            self._last_ckpt_info = \
                {'last_desc': None, 'last_step': 0, 'last_epoch_idx': -1}
            self.create_train_state(
                apply_fn=apply_fn, params=params, optimizer=optimizer, step=0)

        self._default_predictor_fn = partial(
            Predictor,
            deployer=deployer,
            collate_fn=collate_fn,
            params_sharding_rules=params_sharding_rules)

    def create_train_state(
            self, apply_fn, params, optimizer, step, init_opt_state=None):
        params = freeze(params)

        if self._deployer.mesh is None:
            opt_state = init_opt_state if init_opt_state is not None \
                else optimizer.init(params)
            self._state = train_state.TrainState(
                step=step,
                apply_fn=apply_fn,
                params=params,
                tx=optimizer,
                opt_state=opt_state)
            self._state = replicate(self._state)
        else:
            params_spec = self._deployer.get_params_spec(
                params=params,
                params_sharding_rules=self._params_sharding_rules)

            params, opt_state, opt_state_spec = \
                self._deployer.shard_params_and_opt_state(
                    params=params,
                    params_spec=params_spec,
                    optimizer=optimizer,
                    init_opt_state=init_opt_state)

            self._state = train_state.TrainState(
                apply_fn=apply_fn,
                params=params,
                tx=optimizer,
                opt_state=opt_state,
                step=step)

            self._state_spec = train_state.TrainState(
                apply_fn=apply_fn,
                params=params_spec,
                tx=optimizer,
                opt_state=opt_state_spec,
                step=None)

    def setup_running_step(self, dummy_batch):
        train_step_fn = partial(
            default_train_step,
            loss_fn=self._loss_fn,
            lr_schedule_fn=self._lr_schedule_fn,
            under_pmap=(self._deployer.mesh is None))

        eval_step_fn = partial(
            default_eval_step,
            loss_fn=self._loss_fn,
            under_pmap=(self._deployer.mesh is None))

        if self._deployer.mesh is None:
            self._p_train_step = jax.pmap(train_step_fn, axis_name='batch')
            self._p_eval_step = jax.pmap(eval_step_fn, axis_name='batch')
        else:
            data_spec = jax.tree_util.tree_map(
                lambda x: P(*(('dp',) + (None,) * (len(x.shape) - 1))),
                dummy_batch)

            self._p_train_step = pjit(
                train_step_fn,
                in_shardings=(None, self._state_spec, data_spec),
                out_shardings=(self._state_spec, None),
                donate_argnums=(1, ))

            self._p_eval_step = pjit(
                eval_step_fn,
                in_shardings=(self._state_spec, data_spec),
                out_shardings=None)

    def train(self, examples, per_device_batch_size, desc=None):
        data_batches = self._deployer.get_model_input_batches(
            examples=examples,
            per_device_batch_size=per_device_batch_size,
            collate_fn=self._collate_fn,
            shuffle=True,
            shuffle_rng=self._deployer.gen_rng(),
            desc=f'Training ({desc})' if desc is not None else 'Training',
            is_train=True,
            accumulate_grad_batches=self._accumulate_grad_batches)

        for batch in data_batches:
            if self._p_train_step is None:
                self.setup_running_step(dummy_batch=batch)

            train_rng = self._deployer.process_to_run_model(
                self._deployer.gen_rng(), is_prng_key=True)
            self._state, metrics = self._deployer.run_model_step(
                step_fn=self._p_train_step,
                input_args=(train_rng, self._state, batch))

            metrics = self._deployer.process_to_deliver(metrics)
            data_batches.set_postfix(**metrics)
            self._deployer.log_metrics(metrics=metrics, step=self.step)

    def eval_loss(self, examples, per_device_batch_size, desc=None):
        data_batches = self._deployer.get_model_input_batches(
            examples=examples,
            per_device_batch_size=per_device_batch_size,
            collate_fn=self._collate_fn,
            shuffle=False,
            shuffle_rng=None,
            desc=f'Evaluating ({desc})' if desc is not None else 'Evaluating')

        losses = []
        for batch in data_batches:
            if self._p_eval_step is None:
                self.setup_running_step(dummy_batch=batch)

            metrics = self._deployer.run_model_step(
                step_fn=self._p_eval_step, input_args=(self._state, batch))

            metrics = self._deployer.process_to_deliver(metrics)
            data_batches.set_postfix(**metrics)

            losses.append(metrics['loss'].item())

        return np.mean(losses).item()

    def fit(self,
            train_examples,
            per_device_batch_size,
            n_epochs,
            eval_examples=None,
            eval_per_device_batch_size=None,
            eval_loss=True,
            eval_predictor=None,
            eval_metric_fn=None,
            eval_sanity_check=True,
            save_every_ckpt=False,
            save_last_ckpt=False,
            save_argmin_ckpt_by_metrics=None,
            save_argmax_ckpt_by_metrics=None,
            save_opt_states=True):
        if eval_per_device_batch_size is None:
            eval_per_device_batch_size = per_device_batch_size

        if save_argmax_ckpt_by_metrics is None:
            save_argmax_ckpt_by_metrics = []
        if save_argmin_ckpt_by_metrics is None:
            save_argmin_ckpt_by_metrics = []
        min_metrics, max_metrics = {}, {}

        if os.path.exists(f'{self.workdir}/min_metrics.json'):
            min_metrics = json.load(open(
                f'{self.workdir}/min_metrics.json'))
            self._deployer.log_info(min_metrics, title='Detected min_metrics')

        if os.path.exists(f'{self.workdir}/max_metrics.json'):
            max_metrics = json.load(open(
                f'{self.workdir}/max_metrics.json'))
            self._deployer.log_info(max_metrics, title='Detected max_metrics')

        if eval_sanity_check and eval_examples is not None:
            rng_backup = self._deployer._rng
            _, eval_global_batch_size = self._deployer.process_batch_size(
                per_device_batch_size=eval_per_device_batch_size)

            if eval_loss:
                self.eval_loss(
                    examples=eval_examples[:eval_global_batch_size],
                    per_device_batch_size=eval_per_device_batch_size,
                    desc=f'Sanity check')
                self._deployer.log_info(
                    'Sanity check (for evaluation loss) passed.')

            if eval_predictor is not None:
                preds = eval_predictor.predict(
                    examples=eval_examples[:eval_global_batch_size],
                    params=self.params,
                    params_meshed=(self._deployer.mesh is not None),
                    per_device_batch_size=eval_per_device_batch_size,
                    desc=f'Sanity check')
                self._deployer.log_info(
                    'Sanity check (for prediction) passed.')

                if eval_metric_fn is not None:
                    json.dumps(eval_metric_fn(
                        examples=eval_examples[:eval_global_batch_size],
                        preds=preds))
                    self._deployer.log_info(
                        'Sanity check (for evaluation metrics) passed.')

            self._deployer._rng = rng_backup

        for epoch_idx in range(
                self._last_ckpt_info['last_epoch_idx'] + 1, n_epochs):
            if isinstance(train_examples, list):
                epoch_train_examples = train_examples
            else:
                epoch_train_examples = train_examples(epoch_idx=epoch_idx)

            self.train(
                examples=epoch_train_examples,
                per_device_batch_size=per_device_batch_size,
                desc=f'epoch {epoch_idx} / {n_epochs}')

            if eval_examples is None:
                self._deployer.log_info(
                    'No evaluation cuz \'eval_examples\' is None.')
            else:
                eval_metrics = {}

                if eval_loss:
                    loss = self.eval_loss(
                        examples=eval_examples,
                        per_device_batch_size=eval_per_device_batch_size,
                        desc=f'epoch {epoch_idx} / {n_epochs}')
                    eval_metrics['loss'] = loss

                if eval_predictor is not None:
                    preds = eval_predictor.predict(
                        examples=eval_examples,
                        params=self.params,
                        params_meshed=(self._deployer.mesh is not None),
                        per_device_batch_size=eval_per_device_batch_size,
                        desc=f'epoch {epoch_idx} / {n_epochs}')

                    if eval_metric_fn is not None:
                        eval_metrics.update(eval_metric_fn(
                            examples=eval_examples, preds=preds))

                    eval_outputs = [
                        {'example': example, 'pred': pred}
                        for example, pred in zip(eval_examples, preds)]

                    self._deployer.save_outputs(
                        outputs=eval_outputs,
                        desc=f'epoch{epoch_idx}',
                        step=self.step)

                self._deployer.log_info(
                    info=json.dumps(eval_metrics, indent=4),
                    title=f'Eval results',
                    step=self.step)
                self._deployer.log_metrics(metrics={
                    f'eval_{key}': value
                    for key, value in eval_metrics.items()
                }, step=self.step)

                if self.workdir is not None:
                    result_filepath = \
                        f'{self.workdir}/eval_results_epoch{epoch_idx}.json'
                    json.dump(
                        eval_metrics, open(result_filepath, 'w'), indent=4)
                    self._deployer.log_info(
                        f'eval_results saved into {result_filepath}.')

                save_ckpt_kwargs = {
                    'epoch_idx': epoch_idx,
                    'ckpt_dir': f'{self.workdir}/ckpts',
                    'save_opt_state': save_opt_states
                }

                for key in save_argmin_ckpt_by_metrics:
                    assert self.workdir is not None
                    if eval_metrics[key] < min_metrics.get(key, float('inf')):
                        min_metrics[key] = eval_metrics[key]

                        if jax.process_index() == 0:
                            self._deployer.log_info(
                                f'minimal {key} updated to {min_metrics[key]}')
                            json.dump(max_metrics, open(
                                f'{self.workdir}/max_metrics.json', 'w'))

                        self.save_ckpt(desc=f'min_{key}', **save_ckpt_kwargs)

                for key in save_argmax_ckpt_by_metrics:
                    assert self.workdir is not None
                    if eval_metrics[key] > max_metrics.get(key, float('-inf')):
                        max_metrics[key] = eval_metrics[key]

                        if jax.process_index() == 0:
                            self._deployer.log_info(
                                f'maximal {key} updated to {max_metrics[key]}')
                            json.dump(max_metrics, open(
                                f'{self.workdir}/max_metrics.json', 'w'))

                        self.save_ckpt(desc=f'max_{key}', **save_ckpt_kwargs)

            if save_every_ckpt:
                self.save_ckpt(desc=f'epoch_{epoch_idx}', **save_ckpt_kwargs)
            elif save_last_ckpt:
                self.save_ckpt(desc=f'last', **save_ckpt_kwargs)

    def save_ckpt(self, epoch_idx, desc, ckpt_dir, save_opt_state):
        if jax.process_index() == 0:
            os.makedirs(ckpt_dir, exist_ok=True)

        self._deployer.save_params(
            params=self._state.params, ckpt_dir=ckpt_dir, desc=desc)
        self._deployer.save_rng(ckpt_dir=ckpt_dir, desc=desc)

        if save_opt_state:
            self._deployer.save_opt_state(
                opt_state=self._state.opt_state, ckpt_dir=ckpt_dir, desc=desc)

        if jax.process_index() == 0:
            last_ckpt_info = {
                'last_desc': desc,
                'last_step': self.step,
                'last_epoch_idx': epoch_idx
            }
            json.dump(last_ckpt_info, open(
                f'{ckpt_dir}/last_ckpt_info.json', 'w'), indent=4)
            self._deployer.log_info(f'{ckpt_dir}/last_ckpt_info.json updated.')

    def get_default_predictor(self, pred_fn, output_fn=None):
        return self._default_predictor_fn(pred_fn=pred_fn, output_fn=output_fn)

    @property
    def params(self):
        return self._deployer.process_to_deliver(self._state.params)

    @property
    def step(self):
        return self._deployer.process_to_deliver(self._state.step).item()

    @property
    def workdir(self):
        return self._deployer.workdir



#  Copyright 2021 Google LLC
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import numpy as np
import jax


def add_idxes(examples):
    examples_upd = []
    for idx, example in enumerate(examples):
        assert '__idx__' not in example
        # example = copy.deepcopy(example)
        example = {key: example[key] for key in example.keys()}
        example.update({'__idx__': idx})

        examples_upd.append(example)

    return examples_upd


def collate_fn_wrapper(examples, collate_fn):
    idxes = [example.pop('__idx__') for example in examples]
    batch = collate_fn(examples)
    batch['__idx__'] = np.array(idxes)

    return batch


def pred_fn_wrapper(pred_rng, params, batch, pred_fn, under_pmap):
    idxes = batch.pop('__idx__')
    preds = pred_fn(pred_rng=pred_rng, params=params, batch=batch)
    preds = {
        'raw_preds': preds,
        '__idx__': idxes
    }

    if under_pmap:
        return jax.lax.all_gather(preds, axis_name='batch')
    else:
        return preds


def default_output_fn(preds):
    batch_size = jax.tree_util.tree_leaves(preds)[0].shape[0]

    assert jax.tree_util.tree_all(
        jax.tree_util.tree_map(lambda x: x.shape[0] == batch_size, preds))

    outputs = []
    for i in range(batch_size):
        outputs.append(jax.tree_util.tree_map(lambda x: x[i], preds))

    return outputs